/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Sept 2001      */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

#include "oroborus.h"

int mouse_x, mouse_y, lastx=-100, lasty=-100, lastcnt=0, lastbadenter=0, got_button=False;
double lastmv=0.0;
Client *lastc=NULL;
int replay_x, replay_y, replay_if_drag=0;
XButtonEvent replay_event;
Window replay_win;
Time last_button1_time, replay_time;
#ifdef KEYBOARD_SUPPORT
int tab_client_state;
Time last_tab_time;
#endif

#ifdef KEYBOARD_SUPPORT

/*
 * 
 * Handle key press event
 *
 */

void event_key_press(XKeyEvent *ev)
{
  Client *c=NULL;

#ifdef SHOW_TRACE1
  printf("TRACE1: event_handle_key_press\n");
#endif

  c=get_client_from_window(ev->window, WINDOW);

  ev->state=ev->state&(Mod1Mask|ControlMask|ShiftMask);
  
  /* If we have an active client */
  if(c!=NULL)
  {
    /* Maximize client */
    if(ev->keycode==config.keys.maximize.keycode && ev->state==config.keys.maximize.modifier)
      toggle_client_maximized(c);
      
    /* Shade client */
    if(ev->keycode==config.keys.shade.keycode && ev->state==config.keys.shade.modifier)
      toggle_client_shaded(c);

    /* Hide client */
    if(ev->keycode==config.keys.hide.keycode && ev->state==config.keys.hide.modifier)
      minmax_client(c);
     
    /* Close client */ 
    if(ev->keycode==config.keys.close.keycode && ev->state==config.keys.close.modifier) 
      close_client(c); 
       
    /* Move client up */ 
    if(ev->keycode==config.keys.move_up.keycode && ev->state==config.keys.move_up.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x, c->y-10); 
    }
       
    /* Move client down */ 
    if(ev->keycode==config.keys.move_down.keycode && ev->state==config.keys.move_down.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x, c->y+10); 
    }
       
    /* Move client left */ 
    if(ev->keycode==config.keys.move_left.keycode && ev->state==config.keys.move_left.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x-10, c->y); 
    }
       
    /* Move client right */ 
    if(ev->keycode==config.keys.move_right.keycode && ev->state==config.keys.move_right.modifier) 
    {
      mouse_mode=MOUSE_STOPPED;
      move_client(c, c->x+10, c->y); 
    }
 
    if(!(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_MAXIMIZED) && !(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)) 
    { 
      if(ev->keycode==config.keys.resize_up.keycode && ev->state==config.keys.resize_up.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width, c->height-(c->size->flags&PResizeInc?c->size->height_inc:10), False); 
      }
      if(ev->keycode==config.keys.resize_down.keycode && ev->state==config.keys.resize_down.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width, c->height+(c->size->flags&PResizeInc?c->size->height_inc:10), False);
      } 
      if(ev->keycode==config.keys.resize_left.keycode && ev->state==config.keys.resize_left.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width-(c->size->flags&PResizeInc?c->size->width_inc:10), c->height, False);
      } 
      if(ev->keycode==config.keys.resize_right.keycode && ev->state==config.keys.resize_right.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width+(c->size->flags&PResizeInc?c->size->width_inc:10), c->height, False); 
      }
    } 
     
#ifdef LAYER_SUPPORT 
    /* Raise client layer */ 
    if(ev->keycode==config.keys.raise_layer.keycode && ev->state==config.keys.raise_layer.modifier) 
      set_client_layer(c, get_client_layer(c)+1); 
       
    /* Lower client layer */ 
    if(ev->keycode==config.keys.lower_layer.keycode && ev->state==config.keys.lower_layer.modifier) 
      set_client_layer(c, get_client_layer(c)-1); 
#endif /* LAYER */ 
 
#ifdef WORKSPACE_SUPPORT 
    /* Toggle sticky */ 
    if(ev->keycode==config.keys.sticky.keycode && ev->state==config.keys.sticky.modifier) 
      toggle_client_sticky(c); 
       
    /* Move window to next workspace */ 
    if(ev->keycode==config.keys.move_next_ws.keycode && ev->state==config.keys.move_next_ws.modifier) 
      move_client_to_workspace(c, workspace+1); 
 
    /* Move window to prev workspace */ 
    if(ev->keycode==config.keys.move_prev_ws.keycode && ev->state==config.keys.move_prev_ws.modifier) 
      move_client_to_workspace(c, workspace-1); 
       
    /* Move window to specific workspace */ 
    if(ev->keycode==config.keys.move_ws_1.keycode && ev->state==config.keys.move_ws_1.modifier) 
      move_client_to_workspace(c, 0);      
    if(ev->keycode==config.keys.move_ws_2.keycode && ev->state==config.keys.move_ws_2.modifier) 
      move_client_to_workspace(c, 1); 
    if(ev->keycode==config.keys.move_ws_3.keycode && ev->state==config.keys.move_ws_3.modifier) 
      move_client_to_workspace(c, 2); 
    if(ev->keycode==config.keys.move_ws_4.keycode && ev->state==config.keys.move_ws_4.modifier) 
      move_client_to_workspace(c, 3); 
    if(ev->keycode==config.keys.move_ws_5.keycode && ev->state==config.keys.move_ws_5.modifier) 
      move_client_to_workspace(c, 4); 
    if(ev->keycode==config.keys.move_ws_6.keycode && ev->state==config.keys.move_ws_6.modifier) 
      move_client_to_workspace(c, 5); 
    if(ev->keycode==config.keys.move_ws_7.keycode && ev->state==config.keys.move_ws_7.modifier) 
      move_client_to_workspace(c, 6); 
    if(ev->keycode==config.keys.move_ws_8.keycode && ev->state==config.keys.move_ws_8.modifier) 
      move_client_to_workspace(c, 7); 
    if(ev->keycode==config.keys.move_ws_9.keycode && ev->state==config.keys.move_ws_9.modifier) 
      move_client_to_workspace(c, 8); 
#endif /* WORKSPACE */ 
  } 
 
  /* Next window */ 
  if(ev->keycode==config.keys.next_window.keycode && ev->state==config.keys.next_window.modifier) 
  { 
    Client *c2=NULL; 
 
    /* Check time of last tab */ 
    if(ev->time-last_tab_time<TAB_RAISE_TIMEOUT) 
    { 
     /* If window was hidden then hide it again 
     if(tab_client_state==IconicState && c!=NULL) */
     if((tab_client_state!=0) && (c!=NULL) )
        minmax_client(c); 
    } 
    last_tab_time=ev->time; 
    if(c)
      c2=get_next_client(c);
    else if(client_list)
      c2=get_next_client(client_list->prev); 
    if(c2!=NULL) 
    { 
/*      tab_client_state=get_wm_state(c2->window); */
      tab_client_state=c2->is_icon;
      if(tab_client_state!=0)
        minmax_client(c2);
      raise_client(c2); 
      set_focus_client(c2); 
    } 
  }

  /* Previous window */ 
  if(ev->keycode==config.keys.prev_window.keycode && ev->state==config.keys.prev_window.modifier) 
  { 
    Client *c2=NULL; 
 
    /* Check time of last tab */ 
    if(ev->time-last_tab_time<TAB_RAISE_TIMEOUT) 
    { 
     /* If window was hidden then hide it again 
     if(tab_client_state==IconicState && c!=NULL)  */
     if(tab_client_state!=0 && c!=NULL)
        minmax_client(c); 
    } 
    last_tab_time=ev->time; 
    if(c)
      c2=get_prev_client(c);
    else if(client_list)
      c2=get_prev_client(client_list->next); 
    if(c2!=NULL) 
    { 
/*      tab_client_state=get_wm_state(c2->window); */
      tab_client_state=c2->is_icon;
      if(tab_client_state!=0)
        minmax_client(c2);
      raise_client(c2);
      set_focus_client(c2); 
    }
  } 
 
#ifdef WORKSPACE_SUPPORT   
  /* Switch to next workspace */ 
  if(ev->keycode==config.keys.next_ws.keycode && ev->state==config.keys.next_ws.modifier) 
    change_to_workspace(workspace+1); 
 
  /* Switch to prev workspace */ 
  if(ev->keycode==config.keys.prev_ws.keycode && ev->state==config.keys.prev_ws.modifier) 
    change_to_workspace(workspace-1); 
 
  /* Add new workspace */ 
  if(ev->keycode==config.keys.add_ws.keycode && ev->state==config.keys.add_ws.modifier) 
    set_gnome_hint(root, WIN_WORKSPACE_COUNT, config.workspace_count+1); 
 
  /* Delete last workspace */ 
  if(ev->keycode==config.keys.del_ws.keycode && ev->state==config.keys.del_ws.modifier) 
    set_gnome_hint(root, WIN_WORKSPACE_COUNT, config.workspace_count-1); 
 
  /* Switch to specific workspace */ 
  if(ev->keycode==config.keys.ws_1.keycode && ev->state==config.keys.ws_1.modifier) 
    change_to_workspace(0); 
  if(ev->keycode==config.keys.ws_2.keycode && ev->state==config.keys.ws_2.modifier) 
    change_to_workspace(1); 
  if(ev->keycode==config.keys.ws_3.keycode && ev->state==config.keys.ws_3.modifier) 
    change_to_workspace(2); 
  if(ev->keycode==config.keys.ws_4.keycode && ev->state==config.keys.ws_4.modifier) 
    change_to_workspace(3); 
  if(ev->keycode==config.keys.ws_5.keycode && ev->state==config.keys.ws_5.modifier) 
    change_to_workspace(4); 
  if(ev->keycode==config.keys.ws_6.keycode && ev->state==config.keys.ws_6.modifier) 
    change_to_workspace(5); 
  if(ev->keycode==config.keys.ws_7.keycode && ev->state==config.keys.ws_7.modifier) 
    change_to_workspace(6); 
  if(ev->keycode==config.keys.ws_8.keycode && ev->state==config.keys.ws_8.modifier) 
    change_to_workspace(7); 
  if(ev->keycode==config.keys.ws_9.keycode && ev->state==config.keys.ws_9.modifier) 
    change_to_workspace(8); 
#endif /* WORKSPACE */ 
} 
#endif /* KEYBOARD */ 
 
/* 
 * 
 * Handle button press events 
 * 
 */ 

void event_button_press(XButtonEvent *ev) 
{ 
  Client *c=NULL; 
  Client *win=NULL;
  Window ptr_win; 
  int replay=False, focus=True, raise=True, xinfront=True;
  unsigned int stateorg; 
  XWindowChanges wc; 

#ifdef SHOW_TRACE1
  printf("TRACE1: event_button_press (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif
 
  /*got_button=True;*/

/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  if(replay_if_drag==2) {
#ifdef SHOW_EVENTS
    printf("   got event again: replaying...\n");
#endif
    // got click event for the second time, so try to replay it to the window
    // does not work for some reason...?
    XAllowEvents(dpy, ReplayPointer, CurrentTime);
    replay_if_drag=0;
    return;
  } else if(replay_if_drag==1) {
#ifdef SHOW_EVENTS
    printf("   allowing events (sync) on button down...\n");
#endif
    replay_if_drag=0;
    XAllowEvents(dpy, SyncPointer, CurrentTime);
  }
*/

  replay_event=*ev;

  win=get_focus_client();
  if(gX11InFront==0)
    win=NULL;
  else
    xinfront=False;
  
  c=get_client_from_window(ev->window, FRAME); 

  ev->state=ev->state&(Mod1Mask|ControlMask|ShiftMask);
 
  /* Clicked somewhere on a client */ 
  if(c!=NULL) 
  { 
    ptr_win=get_pointer_win(c->frame); 
 
    /* If we want to resize */ 
    if((ptr_win==c->resize && ev->button==Button1 && ev->state==0) || (ev->button==Button2 && ev->state==(Mod1Mask|ControlMask))) 
    { 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime); 
         
      /* Set mouse_x and mouse_y to position of cursor relative to southeast corner of window */ 
      mouse_x=FW-(ev->x_root-FX); 
      mouse_y=FH-(ev->y_root-FY); 
      mouse_mode=MOUSE_RESIZE; 
    } 
 
    /* If we want to maximise */ 
    else if(ptr_win==c->hide && ev->button==Button1 && ev->state==0) 
    { 
      raise=False;
      focus=False;
      xinfront=False;
// This next bit of code makes it behave like OS X windows -i.e. it comes straight in front as soon
// as maximise is pressed, rather than waiting for the user to 'let go' of the button -but I don't like it!!
//      raise_client(c);
//      set_focus_client(c);
//      win=c;
//      raise=True;
//      focus=True;
//      xinfront=true;
      if(c==win)
         c->hide_state=BUTTON_PRESSED;
      if(c!=win)
         c->hide_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_HIDE; 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
      redraw_client_buttons(c); 
    } 
 
    /* If we want to 'iconbar-ise' */ 
    else if(ptr_win==c->icon && ev->button==Button1 && ev->state==0) 
    { 
      if(c==win)
         c->icon_state=BUTTON_PRESSED; 
      if(c!=win)
         c->icon_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_ICON; 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
      raise=False;
      focus=False;
      xinfront=False;
      redraw_client_buttons(c); 
    } 
 
    /* If we want to close */ 
    else if(ptr_win==c->close && ev->button==Button1 && ev->state==0) 
    { 
      if(c==win)
         c->close_state=BUTTON_PRESSED;
      if(c!=win)
         c->close_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_CLOSE; 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
      raise=False;
      focus=False;
      xinfront=False;
      redraw_client_buttons(c); 
    } 
 
    /* If we want to move or windowshade -or if clicking an iconbar */ 
    else if((ptr_win!=c->window && ev->button==Button1 && ev->state==0) || (ev->button==Button1 && ev->state==(Mod1Mask|ControlMask))) 
    { 
      if( c->is_icon!=NOT_ICONBAR )
      {
        // clicked on an iconbar -don't wait for mouse up, just go ahead and restore it
        minmax_client(c);
      }
      else if(ev->time-last_button1_time<DBL_CLICK) 
      { 
        // got a double-click on the title-bar
        toggle_client_shaded(c);
        last_button1_time=0; 
        raise=False;
        focus=False;
        xinfront=False;
      } 
      else
      { 
        XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, move_curs, CurrentTime); 
 
        /* Set mouse_x and mouse_y to position of cursor relative to root window */ 
        mouse_x=ev->x_root; 
        mouse_y=ev->y_root; 
        mouse_mode=MOUSE_MOVE; 
        last_button1_time=ev->time; 
      } 
    } 
 
    /* If we want to 'iconbar-ise' */ 
    else if(ptr_win!=c->window && ev->button==Button2 && ev->state==0) 
    {
      minmax_client(c); 
      raise=False;
      focus=False;
      xinfront=False;
    }

    /* If we want to send to back */ 
    else if((ptr_win!=c->window && ev->button==Button3 && ev->state==0) || (ev->button==Button3 && ev->state==(Mod1Mask|ControlMask))) 
    { 
      lower_client(c); 
      raise=False; 
      focus=False; 
      xinfront=False;
    } 
 
#ifdef WORKSPACE_SUPPORT 
    /* If we want to make window sticky */ 
    else if(ptr_win!=c->window && ev->button==Button2 && ev->state==Mod1Mask) 
      toggle_client_sticky(c); 
#endif /* WORKSPACE */ 
 
#ifdef LAYER_SUPPORT 
    /* If we want to move to a higher layer */ 
    else if(ptr_win!=c->window && ev->button==Button1 && ev->state==Mod1Mask) 
      set_client_layer(c, get_client_layer(c)+1); 
 
    /* If we want to send to a lower layer */ 
    else if(ptr_win!=c->window && ev->button==Button3 && ev->state==Mod1Mask) 
      set_client_layer(c, get_client_layer(c)-1); 
#endif /* LAYER */ 
    else if( gX11InFront==0 ) {
      // click in a window should just bring XDarwin/OroborOSX in front, since it isn't already
//      Bring_To_Front();
/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...
      replay_if_drag=1;
*/
      raise=True;
      focus=True;
      xinfront=True;
    } else if( get_focus_client()!=c ) {
      // click in a window should just bring it in front, since it isn't already
/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...
      replay_if_drag=1;
*/
      raise=True;
      focus=True;
      xinfront=False;
    } else {
      // click in focus window, so allow it through
      replay=True;
      xinfront=False;
    }
 
    if(c->is_icon==NOT_ICONBAR) {
      if(raise) force_raise_client(c); 
      if(xinfront)
        Bring_X_To_Front(1);
      if(raise) raise_client(c); 
      if(focus) set_focus_client(c); 
      if(xinfront)
        Bring_To_Front(0);
    }

    if(replay)  {
#ifdef SHOW_EVENTS
      printf("  allowing events for replay on button down...\n");
#endif
      // Event can be sent through to window
      XAllowEvents(dpy, ReplayPointer, CurrentTime); 
    } else if(replay_if_drag==1) {
//      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
 
    /* Set mouse_x and mouse_y to position of cursor relative to root window */ 
      replay_x=ev->x_root; 
      replay_y=ev->y_root; 
#ifdef SHOW_EVENTS
      printf("  Starting drag check: %d,%d\n",replay_x,replay_y);
#endif
      replay_time=ev->time;
      replay_win=c->frame;
      XAllowEvents(dpy, SyncPointer, CurrentTime);
// try to get around the cursor bug...
      if(xinfront) {
        XUndefineCursor(dpy, c->frame);
        XDefineCursor(dpy, c->frame, normal_curs);
      }
// This is not working, so forget it for now...
      replay_if_drag=0;
    } else {
#ifdef SHOW_EVENTS
      printf("  allowing events (sync) on button down...\n");
#endif
      XAllowEvents(dpy, SyncPointer, CurrentTime);
// try to get around the cursor bug...
      if(xinfront) {
        XUndefineCursor(dpy, c->frame);
        XDefineCursor(dpy, c->frame, normal_curs);
      }
    }
  } 
 
  /* Clicked on root */ 
  else 
  { 
    /* Proxy root click */ 
    c=get_focus_client();
    // don't want any window to have focus...
    gone_to_back();
    XSendEvent(dpy, check_win, False, SubstructureNotifyMask, (XEvent *)ev);
  } 
} 
 
/* 
 * 
 * Handle button release events 
 * 
 */ 

void event_button_release(XButtonEvent *ev) 
{ 
  Client *c=NULL; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_button_release (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus);
#endif 
 
/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  if(replay_if_drag!=0) {
    replay_if_drag=0;
#ifdef SHOW_EVENTS
    printf("   allowing events (sync) on button release...\n");
#endif
  }
*/

  c=get_client_from_window(ev->window, FRAME); 
  if(c!=NULL) 
  { 
    if(c->is_icon!=NOT_ICONBAR)
    {
      /* minmax_client(c); */
    }
    else
    {
      c->hide_state=BUTTON_UNPRESSED; 
      c->icon_state=BUTTON_UNPRESSED;
      c->close_state=BUTTON_UNPRESSED; 
      redraw_client_buttons(c); 
       
      /* Release occured on maximise button */ 
      if(mouse_mode==MOUSE_HIDE || mouse_mode==MOUSE_HIDE_OUT) {
        if(get_pointer_win(c->frame)==c->hide) {
          raise_client(c);
          set_focus_client(c);
          if(gX11InFront==0)
            Bring_To_Front(0);
          toggle_client_maximized(c);
        }
      }
 
      /* Release occured on iconise button */ 
      if(mouse_mode==MOUSE_ICON || mouse_mode==MOUSE_ICON_OUT) {
        if(get_pointer_win(c->frame)==c->icon) 
          minmax_client(c);
      }
 
      /* Release occured on close button */ 
      if(mouse_mode==MOUSE_CLOSE || mouse_mode==MOUSE_CLOSE_OUT) {
        if(get_pointer_win(c->frame)==c->close) 
          close_client(c);
      }
 
      /* If we are resizing */
      if(mouse_mode==MOUSE_RESIZE)
        resize_client(c, c->width, c->height, False); 
    }

  } 
 
  /* Release was on root window */ 
  else 
  { 
    /* Proxy root release */ 
    XSendEvent(dpy, check_win, False, SubstructureNotifyMask, (XEvent *)ev); 
  } 

  XUngrabPointer(dpy, CurrentTime); 
  mouse_mode=MOUSE_NORMAL;
  
  got_button=False;
} 

/*
 * Attempt to sort out a problem that sometimes appears, where the X11 cursor looks like
 * it has frozen (i.e. been grabbed) after a menu item has been chosen that did not work?
 * Don't know why it does this - but this attempt does not fix it either...
 */
 
void force_ungrab()
{
  XAllowEvents(dpy, SyncPointer, CurrentTime);
  XUngrabPointer(dpy, CurrentTime); 
  mouse_mode=MOUSE_NORMAL;
  
  got_button=False;
}

/* 
 * 
 * Handle motion notify events 
 * 
 */ 

void event_motion_notify(XMotionEvent *ev) 
{ 
  Client *c=NULL; 
  Client *win=NULL;
  int height, width, dx, dy, now_x, now_y; 
  XEvent e; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_motion_notify (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  /* If there are other motion events on this window, ignore this one */ 
  if(XCheckTypedWindowEvent(dpy, ev->window, MotionNotify, &e)) 
  { 
    XPutBackEvent(dpy, &e); 
    return; 
  }
/*
  if(!got_button) return;
*/
  win=get_focus_client();
  if(gX11InFront==0)
    win=NULL;

  c=get_client_from_window(ev->window, FRAME); 
  if(c!=NULL) 
  { 
    /* If resizing */ 
    if(mouse_mode==MOUSE_RESIZE) 
    { 
      /* Don't allow resizing if maximized or shaded 
      if(get_gnome_hint(c->window, WIN_STATE)&(WIN_STATE_MAXIMIZED|WIN_STATE_SHADED)) */ 
      if(get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)
        return;
 
      width=(ev->x_root-FX-LW-RW)+mouse_x; 
      height=(ev->y_root-FY-TH-BH)+mouse_y; 
      resize_client(c, width, height, !config.continuous_resize); 
    } 
 
    /* If moving window */ 
    else if(mouse_mode==MOUSE_MOVE) 
    { 
      dx=ev->x_root-mouse_x; 
      dy=ev->y_root-mouse_y; 
      mouse_x=ev->x_root; 
      mouse_y=ev->y_root; 
      move_client(c, c->x + dx, c->y + dy); 
    } 
    else if(mouse_mode==MOUSE_HIDE)
    // must check for enter/leave of window widgets
    {
      if(get_pointer_win(c->frame)!=c->hide) {
        mouse_mode=MOUSE_HIDE_OUT;
        if(c==win)
          c->hide_state=BUTTON_OVER;
        if(c!=win)
          c->hide_state=BUTTON_OVERBACK;  
        redraw_client_buttons(c); }
    }   
    else if(mouse_mode==MOUSE_HIDE_OUT)
    {     
       if(get_pointer_win(c->frame)==c->hide) {
         mouse_mode=MOUSE_HIDE;
         if(c==win)
           c->hide_state=BUTTON_PRESSED;
         if(c!=win)
           c->hide_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    } 
    else if(mouse_mode==MOUSE_ICON)
    {
      if(get_pointer_win(c->frame)!=c->icon) {
        mouse_mode=MOUSE_ICON_OUT;
        if(c==win)
          c->icon_state=BUTTON_OVER;
        if(c!=win)
          c->icon_state=BUTTON_OVERBACK;
        redraw_client_buttons(c); }
    }
    else if(mouse_mode==MOUSE_ICON_OUT)
    {
       if(get_pointer_win(c->frame)==c->icon) {
         mouse_mode=MOUSE_ICON;
         if(c==win)
           c->icon_state=BUTTON_PRESSED;
         if(c!=win)
           c->icon_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    }
    else if(mouse_mode==MOUSE_CLOSE)
    {
      if(get_pointer_win(c->frame)!=c->close) {
        mouse_mode=MOUSE_CLOSE_OUT;
        if(c==win)
          c->close_state=BUTTON_OVER;
        if(c!=win)
          c->close_state=BUTTON_OVERBACK;
        redraw_client_buttons(c); } 
    } 
    else if(mouse_mode==MOUSE_CLOSE_OUT)
    {
       if(get_pointer_win(c->frame)==c->close) {
         mouse_mode=MOUSE_CLOSE;
         if(c==win)
           c->close_state=BUTTON_PRESSED;
         if(c!=win)
           c->close_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    }

  } 
} 
 
/* 
 * 
 * Handle idle time -looks out for cursor over window widgets
 * 
 */ 

void event_idle()
{
  Client *c=NULL;
  Client *win=NULL;
  Window ptrwin=0;
  int now_x, now_y;

#ifdef DEBUG
#ifndef SHOW_IDLE
// Showing all of the debugging junk is a bit much during idle time (background mouse tracking),
// so this makes sure we don't see it unless we specifically ask for it by defining SHOW_IDLE
  return;
#endif
#endif

#ifdef SHOW_TRACE1
  printf("TRACE1: event_idle\n");
#endif

/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  // replaying for drag in background window does not work (for some reason the XAllowEvents with
  // replayPointer does not actually send the event back to the window...)
  if(replay_if_drag==1) {
    // Hmmm... CurrentTime is not actually the value of the current time,
    // so this timer is rubbish anyway...
    if(CurrentTime-replay_time>DBL_CLICK) {
      get_mouse_xy(&now_x,&now_y);
#ifdef SHOW_EVENTS
      printf("  check for drag: %d,%d  (%d,%d)\n",now_x,now_y,replay_x,replay_y);
      printf("  currenttime/replaytime: %#lx,%#lx\n",CurrentTime,replay_time);
#endif
      if(now_x>replay_x+2)
         replay_if_drag=2;
      else if(now_x<replay_x-2)
         replay_if_drag=2;
      else if(now_y>replay_y+2)
         replay_if_drag=2;
      else if(now_y<replay_y-2)
         replay_if_drag=2;
      if(replay_if_drag==2) {
#ifdef SHOW_EVENTS
        printf("  re-sending event for drag...\n");
#endif
        XPutBackEvent(dpy, (XEvent *) &replay_event);
      }
    } else
#ifdef SHOW_EVENTS
      printf("  waiting for DBL_CLICK time...\n");
#endif
    return;
  }
*/

  if(mouse_mode!=MOUSE_NORMAL)
    return;

/*  XGetInputFocus(dpy, &focus_win, &dummy); */

  ptrwin=get_pointer_win(root);
  if( ptrwin!=lastenter ) {
    if( ptrwin!=lastbadenter || lastenter!=0 ) {
#ifdef SHOW_IDLE
      printf("  discrepancy: %#lx %#lx\n",ptrwin,lastenter );
#endif
      lastenter=ptrwin;
      if(lastenter!=0)
        c=get_client_from_window(lastenter, FRAME);
      if(c==NULL) {
#ifdef SHOW_IDLE
        printf("  failed to find client: %#lx %#lx\n",lastenter, root );
#endif
        lastenter=0;
        lastbadenter=ptrwin;
      }
      entered_client(c,True);
    }
  } else if(lastenter!=0)
    c=get_client_from_window(lastenter, FRAME);
  
  ptrwin=0;

  win=get_focus_client();
  if(gX11InFront==0)
    win=NULL;

  if( hacking_focus ) return;

  if(c!=NULL)
  {
    // if over different client, make sure window widgets of last client are reset
    if(lastc!=c && lastc!=NULL) {
#ifdef SHOW_IDLE
       printf("Resetting buttons of last client\n");
#endif
       lastc->hide_state=BUTTON_UNPRESSED;
       lastc->icon_state=BUTTON_UNPRESSED;
       lastc->close_state=BUTTON_UNPRESSED;
       redraw_client_buttons(lastc); }
    lastc=c;
    ptrwin=get_pointer_win(c->frame);
    // check if we're over any of the window widgets
    if(ptrwin==c->hide || ptrwin==c->icon || ptrwin==c->close) {
#ifdef SHOW_IDLE
       printf("Over buttons of current client\n");
#endif
       if(c->hide_state==BUTTON_UNPRESSED || c->icon_state==BUTTON_UNPRESSED || c->close_state==BUTTON_UNPRESSED) {
         // set "x - +" glyphs in all of them appropriately
         if( win==c ) {
           c->hide_state=BUTTON_OVER;
           c->icon_state=BUTTON_OVER;
           c->close_state=BUTTON_OVER;
         } else {
           c->hide_state=BUTTON_OVERBACK;
           c->icon_state=BUTTON_OVERBACK;
           c->close_state=BUTTON_OVERBACK;
         }
#ifdef SHOW_IDLE
         printf("Lighting up buttons of current client\n");
#endif
         redraw_client_buttons(c);
       }
    } else if(c->hide_state!=BUTTON_UNPRESSED || c->icon_state!=BUTTON_UNPRESSED || c->close_state!=BUTTON_UNPRESSED) {
       // one of the widgets had a glyph, so make sure none of them have now
       c->hide_state=BUTTON_UNPRESSED;
       c->icon_state=BUTTON_UNPRESSED;
       c->close_state=BUTTON_UNPRESSED;
#ifdef SHOW_IDLE
       printf("Switching off buttons of current client\n");
#endif
       redraw_client_buttons(c);
       lastc=NULL;
    }
  }
  else if(lastc!=NULL)
  {
    // not over a window any more, so make sure last one has no glyphs in window widgets
    if(lastc->hide_state!=BUTTON_UNPRESSED || lastc->icon_state!=BUTTON_UNPRESSED || lastc->close_state!=BUTTON_UNPRESSED) {
      lastc->hide_state=BUTTON_UNPRESSED;
      lastc->icon_state=BUTTON_UNPRESSED;
      lastc->close_state=BUTTON_UNPRESSED;
#ifdef SHOW_IDLE
       printf("Switching off buttons of last client\n");
#endif
      redraw_client_buttons(lastc);
      lastc=NULL;
    }
  }
}


/* 
 * 
 * Handle enter notify events - and also leave notify events
 * 
 */ 

void event_enter_notify(XCrossingEvent *ev, int entering) 
{ 
  Window ptrwin=0;
  Client *c=NULL; 
  XEvent e; 

#ifdef SHOW_TRACE1 
  printf("TRACE1: event_enter_notify (window,lastenter,entering,lastfocus: %#lx,%#lx,%d,%#lx)\n",ev->window,lastenter,entering,lastfocus); 
#endif
 
  /* If there are other enter/leave notify events, ignore this one */ 
 
  if( entering ) {
    if(XCheckTypedEvent(dpy, EnterNotify, &e))
    {
      XPutBackEvent(dpy, &e);
      return;
    }
    c=get_client_from_window(ev->window,FRAME);
#ifdef DEBUG
    if(c==NULL)
       printf("  got NULL for frame - not changing lastenter\n");
#endif
    if(c!=NULL) {
       lastenter=ev->window;
       entered_client(c,mouse_mode!=MOUSE_STOPPED);
    }
  } else {
    if(XCheckTypedEvent(dpy, LeaveNotify, &e))  
    {
      XPutBackEvent(dpy, &e);
      return;
    }
    if(ev->window==lastenter && lastenter) {
      c=get_client_from_window(ev->window,FRAME);
      if(c!=NULL) {
        ptrwin=get_pointer_win(c->frame);
        if(ptrwin) {
#ifdef DEBUG
          printf("  ignoring leave event: [%#lx], %#lx\n",lastenter,ptrwin);
#endif
          return;
        }
        if(c->is_icon!=NOT_ICONBAR) {
          ptrwin=get_pointer_win(root);
          if(ptrwin) {
            c=get_client_from_window(ptrwin,FRAME);
            if(c!=NULL) {
              if(c->is_icon!=NOT_ICONBAR)
                return;
            }
          }
        }
      }
    }
    lastenter=0;
    entered_client(NULL,mouse_mode!=MOUSE_STOPPED);
  }
#ifdef DEBUG
  printf("  set lastenter to [%#lx] %#lx, [%#lx] %#lx\n",lastenter,ev->window,c==NULL?0:c->frame,c==NULL?0:c->window);
#endif

} 
 
/* 
 * 
 * Handle focus in events 
 * 
 */ 

void event_focus_in(XFocusChangeEvent *ev) 
{ 
  Client *c=NULL; 
  XEvent e; 
 
#ifdef SHOW_TRACE1
  printf("TRACE1: event_focus_in (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif
 
  /* If there are other focus in events, ignore this one */ 
  if(XCheckTypedEvent(dpy, FocusIn, &e)) 
  { 
#ifdef DEBUG
    printf("  ignored...\n");
#endif
    XPutBackEvent(dpy, &e); 
    return; 
  } 
 
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    if(c->is_icon==NOT_ICONBAR && ev->window!=lastfocus) {
      if(config.raise_on_focus || config.focus_style==FOCUS_CLICK) 
        raise_client(c); 
      lastfocus=ev->window;
      redraw_client_frame(c); }
  } 
  else {
    lastfocus=0;
    set_focus_client(get_top_most_client(WIN_LAYER_NORMAL,False)); 
  }
} 
 
/* 
 * 
 * Handle focus out events 
 * 
 */ 

void event_focus_out(XFocusChangeEvent *ev) 
{ 
  Client *c=NULL; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_focus_out (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL)
    redraw_client_frame(c); 
} 
 
/* 
 * 
 * Handle destroy notify events 
 * 
 */  

void event_destroy_notify(XDestroyWindowEvent *ev) 
{ 
  Client *c; 
   
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_destroy_notify (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  if(ev->window==lastfocus)
    lastfocus=0;

  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) { 
    delete_client(c, False); 
  } else if(lastenter) {
    c=get_client_from_window(ev->window, FRAME);
    if(c!=NULL) {
#ifdef DEBUG
      printf("  checking lastenter: [%#lx]",c->frame);
#endif
      if(c->frame==lastenter) {
#ifdef DEBUG
         printf("  got lastenter in destroy: [%#lx]",lastenter);
#endif
         lastenter=0;
      }
    }
  }
} 
 
/* 
 * 
 * Handle unmap notify events 
 * 
 */ 

void event_unmap_notify(XUnmapEvent *ev) 
{ 
  Client *c=NULL; 
 
#ifdef SHOW_TRACE1
  printf("TRACE1: event_unmap_notify (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif
 
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    if(c->ignore_unmap) 
      c->ignore_unmap--; 
    else 
    { 
      delete_client(c, False); 
      set_focus_client(get_top_most_client(WIN_LAYER_NORMAL,False)); 
    } 
  } 
} 
 
/* 
 * 
 * Handle map request events 
 * 
 */ 
void event_map_request(XMapRequestEvent *ev) 
{ 
  Client *c=NULL; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_map_request (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  c=get_client_from_window(ev->window, WINDOW); 
  if(c==NULL) 
    create_new_client(ev->window, True); 
  else 
  { 
    raise_client(c); 
    set_focus_client(c); 
  } 
} 
 
/* 
 * 
 * Handle configure request events 
 * 
 */ 

void event_configure_request(XConfigureRequestEvent *ev) 
{ 
  Client *c=NULL; 
  XWindowChanges wc; 
  XEvent e; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_configure_request (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  /* If there are other configure requests on this window, ignore this one */ 
  if(XCheckTypedWindowEvent(dpy, ev->window, ConfigureRequest, &e)) 
  { 
    XPutBackEvent(dpy, &e); 
    return; 
  } 
 
  wc.x=ev->x; 
  wc.y=ev->y; 
  wc.width=ev->width; 
  wc.height=ev->height; 
  wc.sibling=ev->above; 
  wc.stack_mode=ev->detail; 
  wc.border_width=0; 
 
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
    configure_client(c, ev->value_mask, &wc, c->is_icon!=NOT_ICONBAR); 
  else 
    XConfigureWindow(dpy, ev->window, ev->value_mask, &wc); 
} 
 
/* 
 *  
 * Handle property notify events 
 * 
 */ 

void event_property_notify(XPropertyEvent *ev) 
{ 
  Client *c=NULL; 
  long dummy; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_property_notify (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    switch(ev->atom) 
    { 
    case XA_WM_NAME: 
      if(c->name) XFree(c->name); 
      XFetchName(dpy, c->window, &c->name); 
      XClearWindow(dpy, c->title); 
      redraw_client_title(c); 
      break; 
    case XA_WM_NORMAL_HINTS: 
      XGetWMNormalHints(dpy, c->window, c->size, &dummy); 
      break; 
    } 
  } 
#ifdef WORKSPACE_SUPPORT 
  else 
  { 
    if(ev->atom==gnome[WIN_WORKSPACE_COUNT]) 
      update_workspace_count(); 
  } 
#endif /* WORKSPACE */ 
} 
 
/* 
 * 
 * Handle client message events 
 * 
 */ 

void event_client_message(XClientMessageEvent *ev) 
{ 
  Client *c=NULL; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_client_message (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    /* Client wants to minimise */ 
    if(ev->message_type==icccm[WM_CHANGE_STATE] && ev->format==32 && ev->data.l[0]==IconicState) 
      minmax_client(c); 
 
    /* Client wants to toggle shaded state */ 
    if(ev->message_type==gnome[WIN_STATE] && ev->format==32 && ev->data.l[0]&WIN_STATE_SHADED) 
      toggle_client_shaded(c); 
       
#ifdef WORKSPACE_SUPPORT 
    /* Client wants to toggle sticky state */ 
    if(ev->message_type==gnome[WIN_STATE] && ev->format==32 && ev->data.l[0]&WIN_STATE_STICKY) 
      toggle_client_sticky(c); 
#endif /* WORKSPACE */ 
 
#ifdef LAYER_SUPPORT 
    /* Client wants to change layer */ 
    if(ev->message_type==gnome[WIN_LAYER] && ev->format==32) 
      set_client_layer(c, ev->data.l[0]); 
#endif /* LAYER */ 
  } 
  else if(ev->window==root)  
  { 
#ifdef WORKSPACE_SUPPORT 
    /* Change to new workspace */ 
    if(ev->message_type==gnome[WIN_WORKSPACE] && ev->format==32) 
      change_to_workspace(ev->data.l[0]); 
#endif /* WORKSPACE */ 
  } 
} 
 
#ifdef SHAPE_SUPPORT 
/* 
 * 
 * Handle shape events 
 * 
 */ 

void event_shape(XShapeEvent *ev) 
{ 
  Client *c=NULL; 
  XEvent e; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_shape (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    /* If there are other shape events on this window, ignore this one */ 
    if(XCheckTypedWindowEvent(dpy, ev->window, shape_event, &e)) 
    { 
      XPutBackEvent(dpy, &e); 
      return; 
    } 
 
    set_client_shape(c); 
  } 
} 
#endif /* SHAPE */ 
 
/* 
 * 
 * Main event loop 
 * 
 */ 

void do_event_loop(int sleepus,int trackMouse) 
{ 
int gotev=0;
XEvent ev;

#ifdef SHOW_TRACE1 
  printf("TRACE1: do_event_loop (sleep,moved: %d,%d)\n",sleepus,trackMouse); 
#endif 
 
//    Restart_XDarwin();
    
/*XGetInputFocus(dpy, &lastenter, &dummy);*/
 
    while(XPending(dpy)) 
    { 
      XNextEvent(dpy, &ev); 

      gotev = 1;

      switch(ev.type) 
      { 
#ifdef KEYBOARD_SUPPORT 
      case KeyPress:
#ifdef SHOW_EVENTS
        printf("Got a key press event\n");
#endif
        event_key_press(&ev.xkey); 
        break; 
#endif /* KEYBOARD */ 
      case ButtonPress:
#ifdef SHOW_EVENTS
        printf("Got a button down event\n");
#endif
        event_button_press(&ev.xbutton); 
        break; 
      case ButtonRelease: 
#ifdef SHOW_EVENTS
        printf("Got a button up event\n");
#endif
        event_button_release(&ev.xbutton); 
        break; 
      case MotionNotify: 
#ifdef SHOW_EVENTS
        printf("Got a motion event\n");
#endif
        event_motion_notify(&ev.xmotion); 
        break; 
      case LeaveNotify:
#ifdef SHOW_EVENTS
        printf("Got a leave event\n");
#endif
        event_enter_notify(&ev.xcrossing,False);
        break;
      case EnterNotify: 
#ifdef SHOW_EVENTS
        printf("Got an enter event\n");
#endif
        event_enter_notify(&ev.xcrossing,True);
        break; 
      case FocusIn: 
#ifdef SHOW_EVENTS
        printf("Got a focus in event\n");
#endif
        event_focus_in(&ev.xfocus); 
        break; 
      case FocusOut: 
#ifdef SHOW_EVENTS
        printf("Got a focus out event\n");
#endif
        event_focus_out(&ev.xfocus); 
        break; 
      case DestroyNotify: 
#ifdef SHOW_EVENTS
        printf("Got a destroy event\n");
#endif
        event_destroy_notify(&ev.xdestroywindow); 
        break; 
      case UnmapNotify: 
#ifdef SHOW_EVENTS
        printf("Got an unmap event\n");
#endif
        event_unmap_notify(&ev.xunmap); 
        break; 
      case MapRequest: 
#ifdef SHOW_EVENTS
        printf("Got a map request event\n");
#endif
        event_map_request(&ev.xmaprequest); 
        break; 
      case ConfigureRequest: 
#ifdef SHOW_EVENTS
        printf("Got a configure event\n");
#endif
        event_configure_request(&ev.xconfigurerequest); 
        break; 
      case PropertyNotify: 
        event_property_notify(&ev.xproperty); 
        break; 
      case ClientMessage: 
#ifdef SHOW_EVENTS
        printf("Got a client message event\n");
#endif
        event_client_message(&ev.xclient); 
        break; 
      default:
        gotev = 0;
#ifdef SHAPE_SUPPORT 
        if(ev.type==shape_event) {
          if(shape) {
#ifdef SHOW_EVENTS
            printf("Got a shape event\n");
#endif /* SHOW_EVENTS */
            event_shape((XShapeEvent *)&ev); 
            gotev = 1;
#ifdef SHOW_EVENTS
          } else {
            printf("Got a shape event -but not used...\n");
#endif /* SHOW_EVENTS */
          }
#ifdef SHOW_EVENTS
        } else {
        printf("Got an unknown event, type: %d\n",ev.type);
#endif /* SHOW_EVENTS */
        }
#else
#ifdef SHOW_EVENTS
        printf("Got an unknown event, type: %d\n",ev.type);
#endif /* SHOW_EVENTS */
#endif /* SHAPE */ 
        break; 
      }
    }
    
    // only need to check mouse-over stuff if the mouse has moved
    if(trackMouse!=0 || replay_if_drag!=0)
      event_idle();
      
    gotev = 0;

    if(mouse_mode==MOUSE_STOPPED) mouse_mode=MOUSE_NORMAL;
    
//    Stop_XDarwin();
    
    // oroborus slept for 5 microseconds! -caused far too much CPU usage...
    // This can happily go to several thousand microseconds without noticeable problems
    usleep(sleepus);
    
//    Restart_XDarwin();
} 
