/******************************************/ 
/* Carbon-compatible section of OroborOSX */
/* (c) Adrian Umpleby Sept 2001           */
/*                                        */
/* Original oroborus window manager code  */
/* (c) Ken Lynch Jan 2001                 */
/*                                        */
/* This file started life as main.c from  */
/* Apple's prototype Carbon application   */
/*                                        */
/* Distributed under the GPL              */ 
/* See LICENSE for more details           */ 
/******************************************/ 
 
/* main.c */

#include <stdlib.h> 
#include <stdio.h> 

#include <Carbon/Carbon.h>

#include "main.h"
#include "common.h"

typedef struct{
    char len;
    char thestr[256];
} pstr_t;

WindowRef	lastFrontWin=NULL, noFocusWin=NULL, aboutWin=NULL;
WindowRef	waiting1Win=NULL, waiting2Win=NULL;
char		*Display_Name=NULL, *xpidstr=NULL;
char		myPath[512], randomID[20];
int 		myPathLen, gQuitFlag=0, launchItemsCount=17, reReadOnNextFront=0;
int		gX11InFront=0, on_cmd_line=0, showingAboutWin=0, showingNoFocusWin=0;
int		gHiding_Iconbars=1, gInterleaving_Windows=0, noIdleMax=3, idleCount=0;
int		mouse_mode=MOUSE_NORMAL, real_x, real_y, lastrl_x=0, lastrl_y=0;
int		moving=1, iconbarpos=ICONBARPOS_TOPLEFT, diffscrn_x=0, diffscrn_y=0;
int		lastchk_x=0, lastchk_y=0, lastchk_cnt=0, lastx11_x=0, lastx11_y=0;
long		lastBackTicks, lastIconbarOutTicks=0, iconbar_WaitTicks=100, lastOKTicks=0;
long		lastMovedTicks=0, lastCompareTicks=0;
Boolean		forceOOSXInFront=FALSE, xdarwin_is_stopped=FALSE;
ProcessSerialNumber	xPSN, myPSN;

void Initialize(void);
void EventLoop(void);
void MakeMenu(void);
void DoEvent(EventRecord *event);
void DoIdle(EventRecord *event, int trackMouse);
void DoMenuCommand(long menuResult);
int LaunchMenuItem(char *itemname, int hasext, int hasnl);
void DoAboutBox(void);
void ShowOroborOSXHelp(void);
void DrawWindow(WindowRef window);
int CheckForX(void);
void LaunchX_GetPID(int launchit);
void Bring_To_Front(int quick);
void Bring_X_To_Front(int quick);
void Bring_OOSX_To_Front(void);
static OSErr QuitAppleEventHandler(const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon);
static OSErr ODOCAppleEventHandler(const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon);
static OSErr RAPPAppleEventHandler(const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon);
void Stop_XDarwin();
void Restart_XDarwin();
void Create_Launch_Menu();
void Launch_Startup_Items();

int main(int argc, char *argv[])
{
        int retval, i;

#ifdef SHOW_TRACE1
        printf("TRACE1: main (argc: %d)\n",argc);
#endif

	Initialize();

// Grab the Menu bar from the nib package
	MakeMenu();
        
// Create the special window for the 'all iconbars' case, and the 'no clients' case
        noFocusWin=(WindowRef) MakeWindow("- no window selected -",noFocusWintype);
        waiting1Win=(WindowRef) MakeWindow("Information",waiting1Wintype);

// Check to see if XDarwin is already running
        printf("checking for XDarwin\n");
        if(CheckForX()==0) {
// Try to launch it if not
            waiting2Win=(WindowRef) MakeWindow("Information",waiting2Wintype);
            if(waiting1Win) DisposeWindow(waiting1Win);
            waiting1Win=waiting2Win;
            waiting2Win=NULL;
            LaunchX_GetPID(1);
            retval=-100;
            if(gQuitFlag==0) {
                i=0;
                while(retval!=0 && i<10) {
// See if the window manager can find the display yet
                    waiting2Win=(WindowRef) MakeWindow("Information",waiting3Wintype);
                    if(waiting1Win) DisposeWindow(waiting1Win);
                    waiting1Win=waiting2Win;
                    waiting2Win=NULL;
//                    Restart_XDarwin();
                    retval=orob_main(argc,argv);
                    if(retval!=0) {
                        printf("Waiting for XDarwin to be ready... %d\n",retval);
                        system("/bin/sleep 2");
                        i++;
                    }
                }
            }
            if(retval==0) {
                printf("XDarwin is ready...\n");
            } else {	/* Should have an alert here... */
                fprintf(stderr,"XDarwin has not initialised correctly. Aborting.");
            }
        } else {
// Just get the PID for XDarwin
            LaunchX_GetPID(0);
            waiting2Win=(WindowRef) MakeWindow("Information",waiting3Wintype);
            if(waiting1Win) DisposeWindow(waiting1Win);
            waiting1Win=waiting2Win;
            waiting2Win=NULL;
//            Restart_XDarwin();
            retval=orob_main(argc,argv);
        }
        
        if( retval==0 ) {
            if(waiting1Win) DisposeWindow(waiting1Win);
            EventLoop();
        } else	/* Should have an alert here... */
            fprintf(stderr,"Got an error from orob_main: %d\n",retval);

	return 0;
}

void SetMyPath()
{
    CFURLRef	myURL;
    CFBundleRef mainBundle;

#ifdef SHOW_TRACE2
    printf("TRACE2: SetMyPath\n");
#endif

// Get the main bundle for the app, and find the path
    myPathLen=512;
    mainBundle = CFBundleGetMainBundle();
    myURL=CFBundleCopyBundleURL(mainBundle);
    CFURLGetFileSystemRepresentation(myURL,TRUE,myPath,myPathLen);
    myPathLen=strlen(myPath);
}

void Initialize()	/* Initialize some things */
{
    OSErr	err;

#ifdef SHOW_TRACE1
    printf("TRACE1: Initialize\n");
#endif

// Cursor manager    
    InitCursor();
    
// Make sure path to OroborOSX is set up
    SetMyPath();

// Get the PSN for OroborOSX
    err = GetCurrentProcess(&myPSN);

// Install the quit AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEQuitApplication, NewAEEventHandlerUPP((AEEventHandlerProcPtr)QuitAppleEventHandler), 0, false );
    if (err != noErr)
        ExitToShell();

// Install the rapp AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEReopenApplication, NewAEEventHandlerUPP((AEEventHandlerProcPtr)RAPPAppleEventHandler), 0, false );
    if (err != noErr)
        ExitToShell();

// Install the odoc AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEOpenDocuments, NewAEEventHandlerUPP((AEEventHandlerProcPtr)ODOCAppleEventHandler), 0, false );
    if (err != noErr)
        ExitToShell();
        
// Make a unique (probably!) ID for this process to create temporary files
    sprintf(randomID,"%#lx",random());
}

static OSErr QuitAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
#ifdef SHOW_TRACE1
    printf("TRACE1: QuitAppleEventHandler\n");
#endif

    quit();
    gQuitFlag =  1;
    
    return noErr;
}

static OSErr RAPPAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
#ifdef SHOW_TRACE1
    printf("TRACE1: RAPPAppleEventHandler\n");
#endif

    printf("Got a 'rapp' AppleEvent\n");
    raise_all_clients();
    
    return noErr;
}

static OSErr ODOCAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
    OSType	HLEclass, HLEmsgID;
    OSErr	myErr;
    long	i, ni;
    int		namelen;
    DescType	typeCode;
    Size	dataSize;
    AEDescList	myDesc, newDesc;
    AEKeyword	theAEKeyword;
    Boolean	changed, dummy;
    FSSpec	mySpec;
    AliasHandle	myAlias;
    pstr_t	*pStr;
    char	*cStr;
    
#ifdef SHOW_TRACE1
    printf("TRACE1: ODOCAppleEventHandler\n");
#endif

    HLEmsgID = appleEvt->descriptorType;
    HLEclass = reply->descriptorType;
    
    myErr=AECountItems(appleEvt,&ni);
    
    printf("  Number of initial items: %d\n",ni);
    
    if(ni>0) {
        myErr=AESizeOfNthItem(appleEvt, 1, &typeCode, &dataSize);
        printf("  typeCode,dataSize: %d,%d\n",typeCode,dataSize);
        myErr=AEGetNthDesc(appleEvt,1,typeCode,&theAEKeyword,&myDesc);
        HLEclass=(OSType) theAEKeyword;
        HLEmsgID=myDesc.descriptorType;
        printf("  HLEclass,HLEmsgID: %d,%d\n",HLEclass,HLEmsgID);
        myErr=AECountItems(&myDesc,&ni);
        printf("  Number of actual items: %d\n",ni);
        
        for(i=1;i<=ni;i++) {
            myErr=AESizeOfNthItem(&myDesc,i,&typeCode,&dataSize);
            printf("  item %d   typeCode,dataSize: %d,%d\n",i,typeCode,dataSize);
            myErr=AEGetNthDesc(&myDesc,i,typeCode,&theAEKeyword,&newDesc);
            myAlias=(AliasHandle) newDesc.dataHandle;
            printf("  item %d   myAlias: %#lx\n",i,myAlias);
            myErr=ResolveAlias(NULL,myAlias,&mySpec,&changed);
            if(myErr==noErr) {
// Use the horrible hack to get C string out of a Pascal string
// -there's probably a macro somewhere to do this 'legally'...
                pStr=&(mySpec.name);
                namelen=pStr->len;
                cStr=&(pStr->thestr);
                cStr[namelen]=0;
                printf("  item %d  name: %s\n",i,cStr);
// Now look for that item in the Launch menu, and run the associated script
                myErr=LaunchMenuItem(cStr,1,0);
                if(myErr!=0)
                    printf("OroborOSX: could not launch item %s -got error %d\n",cStr,myErr);
            } else
                printf("  item %d  got an error: %d\n",i,myErr);
        }
    }
    
    return noErr;
}

int CheckForX()
{
    ProcessInfoRec	procInfo;
    ProcessSerialNumber aPSN;
    short		myErr;
    pstr_t		procName;
    
#ifdef SHOW_TRACE1
    printf("TRACE1: CheckForX\n");
#endif

    aPSN.highLongOfPSN=0;
    aPSN.lowLongOfPSN=0;

    myErr=noErr;

// Go through list of applications, looking for XDarwin
    while(myErr==noErr) {
    
        myErr=GetNextProcess(&aPSN);
    
        if(myErr==noErr) {
            procInfo.processName=(char *) &procName;
            procInfo.processAppSpec=NULL;
            myErr=GetProcessInformation(&aPSN,&procInfo);
        } else if(myErr==-600 || myErr==-50)
            myErr=-1000;
        
        if(myErr==-50)
            myErr=noErr;
        else if(myErr==noErr) {
            if(strncmp(procName.thestr,XDARWIN_NAME,XDARWIN_NAMELEN)==0) {
// Found XDarwin!
                xPSN=aPSN;
                return 1;
            }
        }
    }
    
    return 0;
}

void LaunchX_GetPID(int launchit)
{
    char *script=NULL, *lvalue=NULL;
    char buf[20];
    int i;
    FILE *file;

#ifdef SHOW_TRACE1
    printf("TRACE1: LaunchX_GetPID (launchit: %d)\n",launchit);
#endif

#ifdef DEBUG    
    printf("About to malloc script: %d\n",myPathLen);
#endif
    script=malloc(myPathLen+40);
#ifdef DEBUG    
    printf("script: %#lx\n",(long) script);
#endif

    if(script==NULL)
    {
        if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in LaunchX_GetPID.\n", PROGNAME);
            exit(1);
        } else {
            fprintf(stderr, "%s: Memory allocation failed in LaunchX_GetPID.\n", PROGNAME);
            gQuitFlag=-30;
            return;
        }
    }
    
    if(launchit!=0) {
        sprintf(script,"'%s%s", myPath, "/Contents/Resources/startx' &");
        printf("X11 startup script: %s\n",script);
  
// run script to start up X11...
        system(script);
//    system("PATH=.:/usr/bin:/bin:/usr/X11R6/bin:/usr/local/bin:~/bin;startx -- -rootless &");

// Not very Mac-ish way to wait for something...!
        i=0;
        while(CheckForX()==0 && i<15) {
            printf("Waiting for XDarwin to start...\n");
            system("/bin/sleep 2");
            i++;
        }
    
        if(i==15) {
            free(script);
            fprintf(stderr,"%s: failed to start XDarwin",PROGNAME);
            gQuitFlag=-40;
            return;
        }
// wait a couple of seconds to make sure everything has calmed down...
        system("/bin/sleep 2");
    }

// use the random string as part of the PID filename
    sprintf(script,"'%s%s%s", myPath, "/Contents/Resources/getXpid' ",randomID);
    printf("getting XDarwin's pid: %s\n",script);
// run script which finds the PID of XDarwin...
    system(script);
        
// Now open the file (/tmp/XDarwin-pid.'$randomID')
    sprintf(script,"/tmp/XDarwin-pid.%s",randomID);
    if((file=fopen(script, "r"))) 
    {
// and extract the PID from it
        if(fgets(buf, sizeof buf, file)) 
        { 
            lvalue=strtok(buf, "\n"); 
            if(lvalue)
                xpidstr=strdup(lvalue);
            else
                xpidstr=strdup(buf);
            printf("Got XDarwin PID: %s\n",xpidstr);
        } 
        fclose(file);
// and get rid of the file
        sprintf(script,"/bin/rm -f /tmp/XDarwin-pid.%s",randomID);
        printf("Removing file: %s\n",script);
        system(script);
    } else
        printf("  pid file not found: %s\n",script);

    free(script);

// wait a second to make sure everything has calmed down...
    system("/bin/sleep 1");
}

void Restart_XDarwin()
{
    char	script[25];
    
    if(!xpidstr) return;
    
    printf("CHECKING XDARWIN: %d\n",xdarwin_is_stopped==FALSE?0:1);
    if(xdarwin_is_stopped) {
        sprintf(script,"kill -CONT %s",xpidstr);
        printf("STARTING XDARWIN: %s\n",script);
        system(script);
        xdarwin_is_stopped=FALSE;
    }
}

void Stop_XDarwin()
{
    char	script[25];
    
    if(!xpidstr) return;
    
    printf("CHECKING XDARWIN: %d\n",xdarwin_is_stopped==FALSE?0:1);
    if(!xdarwin_is_stopped) {
        sprintf(script,"kill -STOP %s",xpidstr);
        printf("STOPPING XDARWIN: %s\n",script);
        system(script);
        xdarwin_is_stopped=TRUE;
    }
}

void Bring_To_Front(int quick)	/* when an X11 window is clicked, bring everything in front */
{
    OSErr	myErr;
    
#ifdef SHOW_TRACE2
    printf("TRACE2: Bring_To_Front\n");
#endif

//    myErr=SetFrontProcess(&myPSN);
    myErr=SetFrontProcess(&xPSN);
#ifdef DEBUG
    printf("XDarwin has been forced to front\n");
#endif

/*    if(stopx!=0) {
        Stop_XDarwin();
        system("/bin/sleep 1");
        Restart_XDarwin();
    }*/

    if(gX11InFront==0 && quick==0) {
        gX11InFront=1;
        come_to_front();
    }
// Also set a flag so that OroborOSX will be forced in front briefly (in DoIdle),
// so that its menubar will be visible...
    forceOOSXInFront=TRUE;
}

void Bring_OOSX_To_Front()	/* bring OroborOSX in front */
{
    OSErr		myErr;
    
#ifdef SHOW_TRACE2
    printf("TRACE2: Bring_OOSX_To_Front\n");
#endif

    myErr=SetFrontProcess(&myPSN);
//    myErr=SetFrontProcess(&xPSN);
//    gX11InFront=1;
}

void Bring_X_To_Front(int quick)	/* bring XDarwin in front */
{
    OSErr		myErr;
    
#ifdef SHOW_TRACE2
    printf("TRACE2: Bring_X_To_Front\n");
#endif

//    myErr=SetFrontProcess(&myPSN);
    myErr=SetFrontProcess(&xPSN);
#ifdef DEBUG
    printf("XDarwin has been forced to front\n");
#endif
    if(gX11InFront==0 && quick==0) {
        gX11InFront=1;
        come_to_front();
    }
}

void EventLoop()
{
    Boolean	gotEvent;
    EventRecord	event;
    char	script[35];
        
#ifdef SHOW_TRACE1
    printf("TRACE1: EventLoop\n");
#endif

    gQuitFlag = 0;

// Make sure we're all in front once started...
//    Restart_XDarwin();
    Bring_To_Front(0);
//    Stop_XDarwin();

    idleCount=0;
    noIdleMax=3;
// Leave longer gaps between background mouse tracking if tracers are on
#ifdef SHOW_TRACE1
    noIdleMax=10;
#endif
#ifdef SHOW_TRACE2
    noIdleMax=20;
#endif
#ifdef SHOW_TRACE3
    noIdleMax=50;
#endif

// and launch the items in the StartupItems file
    Launch_Startup_Items();

    do
    {
        if(TickCount()>lastOKTicks+600) {
            lastOKTicks=TickCount();
// Write to console for the moment, to try and keep track of when it hangs
            sprintf(script,"echo -n 'reRead: %d    ';date",reReadOnNextFront);
            system(script);
// Do we still need to recheck the Launch Menu Items folder...?
            if(gX11InFront==1) {
                if(reReadOnNextFront<0) {
                    Create_Launch_Menu();
                    reReadOnNextFront=-reReadOnNextFront;
                }
            } else if(reReadOnNextFront>0) {
                Create_Launch_Menu();
                reReadOnNextFront--;
            }
        }

        
//        Stop_XDarwin();
        gotEvent = WaitNextEvent(everyEvent,&event,3,nil);
//        Restart_XDarwin();
        /*
        if(gX11InFront!=0 || mouse_mode!=MOUSE_NORMAL) {
            if( moving==2 )
                gotEvent = WaitNextEvent(everyEvent,&event,2,nil);
            else
                gotEvent = WaitNextEvent(everyEvent,&event,6,nil);
        } else if( moving==2 )
            gotEvent = WaitNextEvent(everyEvent,&event,3,nil);
        else
            gotEvent = WaitNextEvent(everyEvent,&event,8,nil);
        */
        if(gQuitFlag==0) {
            if (gotEvent)
                DoEvent(&event);
            else if(idleCount==noIdleMax) {
                DoIdle(&event,1);
                idleCount=0;
            } else {
                DoIdle(&event,0);
                idleCount++;
            }
        }
        
    } while (gQuitFlag==0);
    
    system("/bin/sleep 1");
    ExitToShell();					
}

long MakeWindow(char *title,int windowtype)
{
    WindowRef	myWindow;
    Handle	menuBar;
    MenuRef	menu;
    long	response;
    OSErr	err;
    OSStatus	myStat;
    IBNibRef	theNIBRef;
    Rect	winrct;
    pstr_t	ptitle;
    
#ifdef SHOW_TRACE3
    printf("TRACE3: MakeWindow (type: %d)\n",windowtype);
#endif

    if(windowtype==x11Wintype || windowtype==noFocusWintype) {
/* Only set title for the x11 windows and the 'no window selected' window */    
        if(title==NULL) return 0;
    
        winrct.left=20;
        winrct.right=21;
        winrct.top=20;
        winrct.bottom=21;
    
        ptitle.len=strlen(title);
        strncpy((char *) &ptitle.thestr,title,ptitle.len);
    }
    
    myStat = CreateNibReference(CFSTR("window"),&theNIBRef);
    
    switch(windowtype)
    {
/* put up the information window */
    case waiting1Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting1"),&myWindow);
        break;
    case waiting2Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting2"),&myWindow);
        break;
    case waiting3Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting3"),&myWindow);
        break;
/* special window for when all X11 windows are iconbars */    
    case noFocusWintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("NoFocusWin"),&myWindow);
        break;
/* put up the about window */
    case aboutWintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("About"),&myWindow);
        break;
/* make a 'dummy' window, so that it appears in dock icon's window list */
    default:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Window"),&myWindow);
        HideWindow(noFocusWin);
        showingNoFocusWin=0;
        break;
    }
    
    if(myWindow==NULL) return 0;
    
/* Only set title if it's an x11 window, or the 'no window selected' window */    
    if(windowtype==x11Wintype || windowtype==noFocusWintype) {
        myStat=SetWindowBounds(myWindow,kWindowContentRgn,&winrct);
        SetWTitle(myWindow,(char *) &ptitle);
    }
    
    ShowWindow(myWindow);
    
#ifdef DEBUG
    printf("Created window: %s  (%#lx)\n",title,myWindow);
#endif

    DisposeNibReference(theNIBRef);

/* make sure it is not on the screen (if it's one of the x11 windows etc.) */    
    if(windowtype==x11Wintype || windowtype==noFocusWintype)
        MoveWindow(myWindow,-5000,-5000,false);
    
    return (long) myWindow;
}

void CheckOSXWinTitle(long winref, char *title)	/* Check title in dock icon window list */
{
    pstr_t	ptitle;

#ifdef SHOW_TRACE3
    printf("TRACE3: CheckOSXWinTitle (winref: %d)\n",winref);
#endif

    if(winref==0) return;
    
    ptitle.len=strlen(title);
    strncpy((char *) &ptitle.thestr,title,ptitle.len);
    
    MoveWindow((WindowRef) winref,-5000,-5000,false);
    SetWTitle((WindowRef) winref, (char *) &ptitle);
}

void DestroyOSXWindow(long winref)
{
    if(winref==0) return;
// This disposes of the OroborOSX window associated with the X11 window
    DisposeWindow((WindowRef) winref);
}

void HideOSXWin(long winref)
{

#ifdef SHOW_TRACE3
    printf("TRACE3: HideOSXWin (winref: %d)\n",winref);
#endif

    if(winref==0) return;
// note: don't actually use this since it removes it from the window list in the dock icon
//    HideWindow((WindowRef) winref);
}

void FocusOSXWin(long winref)
{

#ifdef SHOW_TRACE3
    printf("TRACE3: FocusOSXWin (winref: %d)\n",winref);
#endif

    if(winref==0) {
        if(showingAboutWin) {
            ShowWindow(aboutWin);
            SelectWindow(aboutWin);
            HideWindow(noFocusWin);
            lastFrontWin=aboutWin;
            showingNoFocusWin=1;
#ifdef DEBUG
            printf("Selected window: %#lx\n",aboutWin);
#endif
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0);
        } else {
            ShowWindow(noFocusWin);
            showingNoFocusWin=1;
            MoveWindow(noFocusWin,-5000,-5000,false);
            SelectWindow(noFocusWin);
#ifdef DEBUG
            printf("Selected window: %#lx\n",noFocusWin);
#endif
            lastFrontWin=noFocusWin;
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0);
        }
        return;
    }
    
    HideWindow(noFocusWin);
    showingNoFocusWin=0;
    
    ShowWindow((WindowRef) winref);
    SelectWindow((WindowRef) winref);
    lastFrontWin=(WindowRef) winref;
#ifdef DEBUG
    printf("Selected window: %#lx\n",winref);
#endif
}

void will_hide_all_iconbars()
{
    lastIconbarOutTicks=TickCount();
#ifdef SHOW_TRACE3
    printf("TRACE3: will_hide_all_iconbars (lastIconbarOutTicks: %d)\n",lastIconbarOutTicks);
#endif
}

/*
 * Rather 'non-Macish' way to get a list of files in the Launch Menu Items folder,
 * and then use it to create the user items for the launch menu
 */
void Create_Launch_Menu()
{
    MenuRef	menu;
    int		i, buflen;
    char	*script, *buf;
    FILE	*file;
    
    menu = GetMenuHandle( menuLaunch );

#ifdef SHOW_TRACE1    
    printf("TRACE1: Create_Launch_Menu\n");
#endif

#ifdef DEBUG    
    printf("About to malloc script & buf: %d\n",myPathLen);
#endif
    script=malloc(myPathLen+50);
    buf=malloc(256);
#ifdef DEBUG    
    printf("script,buf: %#lx,%#lx\n",(long) script, (long) buf);
#endif

    if(script==NULL || buf==NULL)
    {
        if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Create_Launch_Menu.\n", PROGNAME);
            exit(1);
        } else {
// need a nice alert here...
            fprintf(stderr, "%s: Memory allocation failed in Create_Launch_Menu.\n", PROGNAME);
            gQuitFlag=-35;
            return;
        }
    }
    
    sprintf(script,"'%s/Contents/Resources/launchlist' %s",myPath,randomID);
    printf("launchlist script: %s\n",script);
// run script to get list of items...
    system(script);

// This is the file that the script should have just created
    sprintf(script,"/tmp/launchlist.%s",randomID);
#ifdef DEBUG
    printf("file to open: '%s'\n",script);
#endif

// Now check to see if the file exists
    if((file=fopen(script, "r"))) 
    {
// Now that we know the file exists, we start fiddling the menu entries
        launchItemsCount=CountMenuItems(menu);
        for(i=itemUserItems;i<launchItemsCount;i++) {
            DeleteMenuItem(menu,itemUserItems+1); }
        launchItemsCount=itemUserItems;
        
// hack to easily convert the string to a pascal string
        buf++;
        
// now extract the lines from file (max 254 characters)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
                printf("Short line: (%d) %s",buflen,buf);
            } else if(strncmp(buf,"#",1)) {
                printf("Got an item: %s  ",buf);
                // now hack it to a pascal string
                buf--;
                buf[0]=buflen;
                AppendMenu(menu,buf);
                buf++;
                launchItemsCount++;
            } else
                printf("Comment: %s",buf);
        } 
// close the file and get rid of it
        fclose(file);
        sprintf(script,"/bin/rm -f /tmp/launchlist.%s",randomID);
        printf("Removing file: %s\n",script);
        system(script);
    } else
        printf("  launchlist file not found: '%s'\n",script);
    
    free(script);
    
// make sure that we undo the hack before freeing this...
    buf--;
    free(buf);

// and make sure we have the right number of menu items
    launchItemsCount=CountMenuItems(menu);
}

void MakeMenu()		/* Put up a menu */
{
    Handle	menuBar;
    MenuRef	menu;
    long	response;
    OSErr	err;
    OSStatus	myStat;
    IBNibRef	theNIBRef;
    
#ifdef SHOW_TRACE1
    printf("TRACE1: MakeMenu\n");
#endif

    myStat = CreateNibReference(CFSTR("menubar"),&theNIBRef);
    
    myStat = SetMenuBarFromNib(theNIBRef,CFSTR("MainMenu"));
        
    DisposeNibReference(theNIBRef);

    gHiding_Iconbars=1;
    gInterleaving_Windows=0;
    
    iconbarpos=ICONBARPOS_TOPLEFT;
    Setup_Options_Menu(itemTopLeft);
    Setup_Options_Menu(itemMedIconbars);
    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0);
    Create_Launch_Menu();
}

void Setup_Options_Menu(int which)
{
    MenuRef	menu;
    
#ifdef SHOW_TRACE3
    printf("TRACE3: Setup_Options_Menu (which: %d)\n",which);
#endif

    menu = GetMenuHandle( menuOptions );
    
    if(gHiding_Iconbars==0)
        CheckMenuItem( menu, itemHideIconbars, false );
    else
        CheckMenuItem( menu, itemHideIconbars, true );
        
    if(gInterleaving_Windows==0)
        CheckMenuItem( menu, itemInterleave, false );
    else
        CheckMenuItem( menu, itemInterleave, true );
    
    if(which>=itemSlowIconbars) {
        CheckMenuItem( menu, itemSlowIconbars, false );
        CheckMenuItem( menu, itemMedIconbars, false );
        CheckMenuItem( menu, itemFastIconbars, false );
        SetItemMark( menu, which, diamondMark );
    } else if(which>0) {
        CheckMenuItem( menu, itemTopLeft, false );
        CheckMenuItem( menu, itemTopRight, false );
        CheckMenuItem( menu, itemBotLeft, false );
        CheckMenuItem( menu, itemBotRight, false );
        SetItemMark( menu, which, diamondMark );
    }
}

void Compare_Screens(int scrnx, int scrny, int x11x, int x11y)
{
    Point	mspt;
    long	nowTicks, notMovedTicks, notCompareTicks;
    
    if(gX11InFront==0) return;
// once we have a high enough confidence, don't check any more...
    if(lastchk_cnt>1) return;

#ifdef SHOW_TRACE3
    printf("TRACE3: Compare_Screens\n");
#endif

    GetGlobalMouse(&mspt);
    real_x=mspt.h-diffscrn_x;
    real_y=mspt.v-diffscrn_y;
    
    moving=0;
    if(real_x!=scrnx)
        moving=1;
    else if(real_y!=scrny)
        moving=1;
        
//    printf("Mouse co-ords: %d,%d  %d,%d  %d,%d\n",real_x,real_y,scrnx,scrny,x11x,x11y);
    
    if(moving!=0) {
#ifdef SHOW_MOUSE
        printf("Mouse moved: %d,%d  (from %d,%d)\n",real_x,real_y,lastrl_x,lastrl_y);
#endif
        lastMovedTicks=TickCount();
        lastrl_x=real_x;
        lastrl_y=real_y;
        return;
    }

// Hmmmm... XDarwin 1.0a2 reports funny co-ords once we get inside the menu bar, so ignore points near there
    if(mspt.v>-4 && mspt.v<25) return;

    nowTicks=TickCount();
    
    if(lastCompareTicks==0) {
        lastCompareTicks=nowTicks;
        return;
    }
    if(lastMovedTicks==0) {
        lastMovedTicks=nowTicks;
        return;
    }
    notMovedTicks=nowTicks-lastMovedTicks;
    notCompareTicks=nowTicks-lastCompareTicks;
    
// leave a gap of at least 5 ticks between each compare
    if(notCompareTicks<5) return;
// make a note of the time when we got here
    lastCompareTicks=nowTicks;

// and only compare if we have definitely been here before, since the cursor stopped moving...
    if(notCompareTicks*2>notMovedTicks) return;
    
// only compare if the cursor has been still for at least sixth of a second...
//    if(notMovedTicks<10) return;
// ...and if it is not in the same place it was at the last check...
//    if(lastchk_x==mspt.h && lastchk_y==mspt.v) return;
// ...and if the x11 pointer does not appear to have frozen
//    if(lastx11_x==x11x && lastx11_y==x11y) return;

    lastrl_x=x11x;
    lastrl_y=x11y;

// now reset the deltas...
    diffscrn_x += scrnx-x11x;
    diffscrn_y += scrny-x11y;
    
    if(scrnx!=x11x || scrny!=x11y) {
// start again with the confidence factor
        lastchk_cnt=0;
#ifdef SHOW_SCREENDIFF
        printf("New screen difference: deltax=%d, deltay=%d   ",diffscrn_x,diffscrn_y);
        printf("(reported coords: %d,%d  %d,%d  %d,%d)\n",mspt.h,mspt.v,scrnx,scrny,x11x,x11y);
#endif
    } else {
// got the same as at the last check, so increment the confidence factor
        lastchk_cnt++;
#ifdef SHOW_SCREENDIFF
        printf("Screen compare confidence: %d    deltax=%d, deltay=%d\n",lastchk_cnt,diffscrn_x,diffscrn_y);
#endif
    }
    
    lastchk_y=mspt.v;
    lastchk_x=mspt.h;
    
    lastx11_x=x11x;
    lastx11_y=x11y;
    
    real_x=x11x;
    real_y=x11y;
}

void Setup_Window_Menu(int has_frame, int has_close, int has_min, int has_zoom, int is_float, int can_af, int can_ar)
{
    MenuRef	menu;
    
#ifdef SHOW_TRACE3
    printf("TRACE3: Setup_Window_Menu (has_frame,can_af,can_ar: %d,%d,%d)\n",has_frame,can_af,can_ar);
#endif

    menu = GetMenuHandle( menuWindow );
    
    if(has_frame==0) {
        EnableMenuItem(menu,0);
        DisableMenuItem(menu,itemWindowshade);
        DisableMenuItem(menu,itemMinimise);
        DisableMenuItem(menu,itemZoom);
        DisableMenuItem(menu,itemFloat);
        DisableMenuItem(menu,itemAutoFocus);
        DisableMenuItem(menu,itemAutoRaise);
        DisableMenuItem(menu,itemClose);
//        DisableMenuItem(menu,0);
        DrawMenuBar();
        return;
    } else
        EnableMenuItem(menu,0);

    EnableMenuItem(menu,itemWindowshade);
    EnableMenuItem(menu,itemMinimise);
    EnableMenuItem(menu,itemZoom);
    EnableMenuItem(menu,itemFloat);
    EnableMenuItem(menu,itemAutoFocus);
    EnableMenuItem(menu,itemAutoRaise);
    EnableMenuItem(menu,itemClose);

    if(has_close==0)
        DisableMenuItem(menu,itemClose);
    
    if(has_min==0)
        DisableMenuItem(menu,itemMinimise);

    if(has_zoom==0)
        DisableMenuItem(menu,itemZoom);

    if(is_float==0)
        DisableMenuItem(menu,itemFloat);
    else if(is_float==1)
        CheckMenuItem(menu,itemFloat,true);
    else if(is_float==-1)
        CheckMenuItem(menu,itemFloat,false);
        
    if(can_af==0)
        DisableMenuItem(menu,itemAutoFocus);
        
    if(can_ar==0)
        DisableMenuItem(menu,itemAutoRaise);

    DrawMenuBar();
}

void DoIdle(EventRecord *event, int trackMouse)
{
    short		myErr;
    Point		mspt;
    ProcessSerialNumber	frontPSN;
    Boolean		x11InFront;
    Boolean		same, cursorInMenubar=FALSE;
    WindowRef		currentFrontWin;

#ifdef DEBUG
    ProcessInfoRec	procInfo;
    pstr_t		myName;
    pstr_t		procName;
#endif

#ifdef SHOW_TRACE1
    printf("TRACE1: DoIdle (gX11InFront: %d)\n",gX11InFront);
#endif

    myErr=GetFrontProcess(&frontPSN);
    if(myErr!=noErr)
        return;
        
    myErr = GetCurrentProcess(&myPSN);
    if(myErr!=noErr)
        return;
        
#ifdef DEBUG
    procInfo.processAppSpec=NULL;
    procInfo.processName=(char *) &myName;
    myErr=GetProcessInformation(&myPSN,&procInfo);
    if(myErr!=noErr)
        return;
    myName.thestr[(int) myName.len]=0;

    procInfo.processAppSpec=NULL;
    procInfo.processName=(char *) &procName;
    myErr=GetProcessInformation(&frontPSN,&procInfo);
    if(myErr!=noErr)
        return;
    procName.thestr[(int) procName.len]=0;
#endif

    myErr=SameProcess(&xPSN,&frontPSN,&x11InFront);
    if(myErr!=noErr)
        return;

    if(x11InFront)
        xPSN=frontPSN;
        
    myErr=SameProcess(&myPSN,&frontPSN,&same);
    if(myErr!=noErr)
        return;

    GetGlobalMouse(&mspt);
    cursorInMenubar=FALSE;
    if(x11InFront || same)
        if(mspt.h>=0 && mspt.v>=0 && mspt.v<22 && mspt.h<=800)
            cursorInMenubar=TRUE;
    
    currentFrontWin=FrontWindow();

    if(x11InFront) {
#ifdef SHOW_FRONT
        printf("FRONT: XDarwin\n");
#endif
// Make sure OroborOSX application is in front if the about window is frontmost...
        if(currentFrontWin==aboutWin) {
            Bring_OOSX_To_Front();
            same=TRUE;
            x11InFront=FALSE;
            forceOOSXInFront=FALSE;
// Also needs to be frontmost if the cursor is in the menubar - but we still want to
// act as if XDarwin is in front (i.e. keep the current X11 focus window)
        } else if(cursorInMenubar) {
            Bring_OOSX_To_Front();
            forceOOSXInFront=FALSE;
        } else if(forceOOSXInFront) {
            Bring_OOSX_To_Front();
            forceOOSXInFront=FALSE;
        }
    } else if(same) {
#ifdef SHOW_FRONT
        printf("FRONT: OroborOSX\n");
#endif
// But if the OroborOSX application is in front, and the about window is not frontmost...
        if(currentFrontWin!=aboutWin) {
// ...then we must make sure the right window has the focus (before XDarwin comes in front)
            if(currentFrontWin!=lastFrontWin) {
                MoveWindow(currentFrontWin,-5000,-5000,false);
#ifdef SHOW_EVENTS
                printf("New window in front: %#lx %#lx\n",currentFrontWin,lastFrontWin);
#endif
// tell the window manager that a new X11 window has come in front
//		Restart_XDarwin();
                change_focus((long) currentFrontWin);
//		Stop_XDarwin();
                lastFrontWin=currentFrontWin;
            }
// ...and then we want XDarwin in front
// (unless the cursor is in the menubar, in which case we pretend it is)
            if(!cursorInMenubar) {
//		Restart_XDarwin();
                Bring_X_To_Front(0);
//            	Stop_XDarwin();
            } else if(forceOOSXInFront)
                forceOOSXInFront=FALSE;
            same=FALSE;
            x11InFront=TRUE;
        }
#ifdef SHOW_FRONT
    } else {
        printf("FRONT: unknown\n");
#endif
    }

// Check to see if the frontmost window has changed
    if(currentFrontWin!=lastFrontWin) {
        if(currentFrontWin==aboutWin) {
// Disable the window menu if the about window is frontmost
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0);
        } else {
// And now make sure the frontmost window is off the screen
            MoveWindow(currentFrontWin,-5000,-5000,false);
#ifdef SHOW_EVENTS
            printf("New window in front: %#lx %#lx\n",currentFrontWin,lastFrontWin);
#endif
            if(x11InFront) {
// tell the window manager that a new X11 window has come in front
//                Restart_XDarwin();
                change_focus((long) currentFrontWin);
//                Stop_XDarwin();
            }
        }
        lastFrontWin=currentFrontWin;
    }
        
    real_x=mspt.h-diffscrn_x;
    real_y=mspt.v-diffscrn_y;
    
    moving=0;
    if(real_x>lastrl_x+3)
        moving=2;
    else if(real_y>lastrl_y+3)
        moving=2;
    else if(real_x<lastrl_x-3)
        moving=2;
    else if(real_y<lastrl_y-3)
        moving=2;
    else if(real_x!=lastrl_x)
        moving=1;
    else if(real_y!=lastrl_y)
        moving=1;
        
    if(moving!=0) {
#ifdef SHOW_MOUSE
        printf("Mouse moved: %d,%d  (from %d,%d)\n",real_x,real_y,lastrl_x,lastrl_y);
#endif
        lastMovedTicks=TickCount();
        lastrl_x=real_x;
        lastrl_y=real_y;
    } else if(TickCount()>lastMovedTicks+10)
        moving=10;


    if(x11InFront) {
    
	if(gX11InFront==0) {
#ifdef DEBUG
            printf("%s has come to front\n",procName.thestr);
#endif
            gX11InFront=1;
//            Restart_XDarwin();
            come_to_front();
//            Stop_XDarwin();
        }

        if( moving==0 )
            do_event_loop(1000,0);
        else
            do_event_loop(500,1);

    } else {
    
        if(gX11InFront!=0) {
#ifdef DEBUG
            if(same)
                printf("%s: I am in front!\n",procName.thestr);
            else
                printf("%s has come to front\n",procName.thestr);
#endif
            gX11InFront=0;
//            Restart_XDarwin();
            gone_to_back();
//            Stop_XDarwin();
        }
        
        if( moving==0 )
            do_event_loop(2000,0);
        else
            do_event_loop(1000,1);
    }
    
    if(gQuitFlag==0)
        if(lastIconbarOutTicks!=0)
            if(TickCount()>lastIconbarOutTicks+iconbar_WaitTicks) {
//                Restart_XDarwin();
                hide_all_iconbars();
//            	Stop_XDarwin();
            }
}

void DoEvent(EventRecord *event)
{
    short	part;
    Boolean	hit;
    char	key;
    Rect	tempRect;
    WindowRef	whichWindow;
        
#ifdef SHOW_TRACE1
    printf("TRACE1: DoEvent (type: %d)\n",event->what);
#endif

    switch (event->what) 
    {
        case mouseDown:
#ifdef SHOW_EVENTS
            printf("Got mouse down...\n");
#endif
            part = FindWindow(event->where, &whichWindow);
            switch (part)
            {
                case inMenuBar:  /* process a moused menu command */
                    DoMenuCommand(MenuSelect(event->where));
                    break;
                    
                case inSysWindow:
                    break;
                
                case inContent:
                    if (whichWindow != FrontWindow()) 
                        SelectWindow(whichWindow);
                    break;
                
                case inDrag:	/* pass screenBits.bounds */
//                    Restart_XDarwin();
                    if(showingAboutWin)
                        if(whichWindow==aboutWin)
                            if(gX11InFront==1) {
                                gX11InFront=0;
                                gone_to_back();
                                }
                    GetRegionBounds(GetGrayRgn(), &tempRect);
                    DragWindow(whichWindow, event->where, &tempRect);
//                    Stop_XDarwin();
                    break;
                    
                case inGrow:
                    break;
                    
                case inGoAway:
//                    Restart_XDarwin();
                    hit = TrackBox(whichWindow, event->where, part);
//                    Stop_XDarwin();
                    if (hit) 
                    {
                        if(whichWindow==aboutWin) {
#ifdef DEBUG
                            printf("Hiding about window\n");
#endif
                            showingAboutWin=0;
                            DisposeWindow(aboutWin);
                            aboutWin=NULL;
                            if(showingNoFocusWin!=0) {
                                ShowWindow(noFocusWin);
                                SelectWindow(noFocusWin);
                            }
                        } else {
#ifdef DEBUG
                            printf("Closing window: %#lx (%#lx)\n",whichWindow,aboutWin);
#endif
                            DisposeWindow(whichWindow); }
/*                        quit();
                        system("/bin/sleep 1");
                        ExitToShell();*/
                    }
                    break;
                    
                case inZoomIn:
                case inZoomOut:
//                    Restart_XDarwin();
                    hit = TrackBox(whichWindow, event->where, part);
//                    Stop_XDarwin();
                    if (hit) 
                    {
                        SetPort(GetWindowPort(whichWindow));   // window must be current port
                        EraseRect(GetWindowPortBounds(whichWindow, &tempRect));   // inval/erase because of ZoomWindow bug
                        ZoomWindow(whichWindow, part, true);
                        InvalWindowRect(whichWindow, GetWindowPortBounds(whichWindow, &tempRect));	
                    }
                    break;
                }
                break;
		
                case keyDown:
		case autoKey:
                    key = event->message & charCodeMask;
                    if (event->modifiers & cmdKey)
                        if (event->what == keyDown)
                            DoMenuCommand(MenuKey(key));
		case activateEvt:	       /* if you needed to do something special */
                    break;
                    
                case updateEvt:
			DrawWindow((WindowRef) event->message);
			break;
                        
                case kHighLevelEvent:
			AEProcessAppleEvent( event );
			break;
		
                case diskEvt:
			break;
                
                default:
                        
                        break;
	}
}

void RunX11App(char *path, int pathlen, char *name, int namelen, char *args, int arglen)
{
    char *script=NULL;

#ifdef SHOW_TRACE3
    printf("TRACE3: RunX11App (pathlen: %d)\n",pathlen);
#endif

#ifdef DEBUG
    printf("About to malloc script in RunX11App: %d %d\n",myPathLen,namelen);
#endif
    script=malloc(myPathLen+namelen+arglen+strlen(Display_Name)+50);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif

    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in RunX11App.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in RunX11App.\n", PROGNAME);
            return;
        }
    }
    
    if (arglen==0)
        sprintf(script,"'%s%s%s%s%s%s", myPath, "/Contents/LaunchScripts/generic' '",Display_Name,"' '",name,"'");
    else
        sprintf(script,"'%s%s%s%s%s%s%s%s", myPath, "/Contents/LaunchScripts/generic' '",Display_Name,"' '",name,"' """,args,"""");

#ifdef DEBUG
    printf("X11 Launch script: %s\n",script);
#endif

// run script to launch X11 application...
    system(script);
    
    free(script);
}

void RunX11Script(char *path, int pathlen, int item)
{
    MenuRef	menu;
    Str255	itemStr;
    int		namelen;
    pstr_t	*pStr;
    char	*cStr;
    char *script=NULL;

#ifdef SHOW_TRACE3
    printf("TRACE3: RunX11Script (pathlen,item: %d,%d)\n",pathlen,item);
#endif

    menu = GetMenuHandle( menuLaunch );
    GetMenuItemText(menu,item,itemStr);
    
// hack to get the string out of the pascal string
// - there's probably a macro to do this 'legally' somewhere!
    pStr=&itemStr;
    namelen=pStr->len;
    cStr=&(pStr->thestr);
    cStr[namelen]=0;
    
#ifdef DEBUG
    printf("Got item: (%d) %s\n",namelen,cStr);    
    printf("About to malloc script in RunX11Script: %d %d\n",myPathLen,namelen);
#endif
    script=malloc(myPathLen+namelen+strlen(Display_Name)+40);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in RunX11Script.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in RunX11Script.\n", PROGNAME);
            return;
        }
    }
    
    sprintf(script,"'%s%s%s%s%s'",myPath,"/../Launch Menu Items/",cStr,".x11app' '",Display_Name);

#ifdef DEBUG
    printf("Running Launch Item: %s\n",script);
#endif
    system(script);
    
    free(script);
}

void Edit_Startup_Items()
{
    char *script=NULL;

#ifdef SHOW_TRACE1
    printf("TRACE1: Edit_Startup_Items (pathlen: %d)\n",pathlen);
#endif

#ifdef DEBUG
    printf("About to malloc script in Edit_Startup_Items: %d %d\n",myPathLen,namelen);
#endif
    script=malloc(myPathLen+65);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Edit_Startup_Items.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Edit_Startup_Items.\n", PROGNAME);
            return;
        }
    }

// crude editing method - just open up the file in TextEdit for the user to fiddle with...
    sprintf(script,"/usr/bin/open -a TextEdit '%s%s'",myPath,"/../Launch Menu Items/StartupItems");
    printf("%s\n",script);
    system(script);
    
    free(script);
}

void Launch_Startup_Items()
{
    char *script=NULL;
    char buf[255];
    FILE *file;
    int buflen;

#ifdef SHOW_TRACE1
    printf("TRACE1: Launch_Startup_Items (pathlen: %d)\n",pathlen);
#endif

#ifdef DEBUG
    printf("About to malloc script in Launch_Startup_Items: %d %d\n",myPathLen,namelen);
#endif
    script=malloc(myPathLen+40);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Launch_Startup_Items.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Launch_Startup_Items.\n", PROGNAME);
            return;
        }
    }

// crude editing method - just open up the file in TextEdit for the user to fiddle with...
    sprintf(script,"%s%s",myPath,"/../Launch Menu Items/StartupItems");
    printf("StartupItems file: %s\n",script);
    
// Now check to see if the file exists
    if((file=fopen(script, "r"))) 
    {        
// and extract each line from the file (max 254 characters for a line)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
#ifdef DEBUG
                printf("Short line: %s  (%d)",buf,buflen);
#endif
            } else if(strncmp(buf,"#",1)) {
// Now try to launch it (must be the same as an item in the Launch menu)
                printf("Got an item: %s  (%d)",buf,buflen);
                LaunchMenuItem(buf,0,1);
#ifdef DEBUG
            } else {
                printf("Comment: %s",buf);
#endif
            }
        } 
// close the file
        fclose(file);
    } else
        printf("  StartupItems file not found: '%s'\n",script);
    
    free(script);
}

void Edit_LaunchMenu_Items()
{
    char *script=NULL;

#ifdef SHOW_TRACE1
    printf("TRACE1: Edit_LaunchMenu_Items (pathlen: %d)\n",pathlen);
#endif

#ifdef DEBUG
    printf("About to malloc script in Edit_LaunchMenu_Items: %d %d\n",myPathLen,namelen);
#endif
    script=malloc(myPathLen+45);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Edit_LaunchMenu_Items.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Edit_LaunchMenu_Items.\n", PROGNAME);
            return;
        }
    }

// crude editing method - just open up the Launch Menu Items folder for the user to fiddle with...
    sprintf(script,"/usr/bin/open '%s%s'",myPath,"/../Launch Menu Items/");
    printf("%s\n",script);
    system(script);
    
    free(script);

// Once OroborOSX is brought in front, it will recheck the Launch Menu Items folder
// and continue to do so every ten seconds for about a minute - hopefully this will
// allow it to catch any changes that are made to the items in the folder...
    reReadOnNextFront=-6;
    lastOKTicks=TickCount();
}

void LaunchItem(int item)
{

#ifdef SHOW_TRACE2
    printf("TRACE2: LaunchItem (item: %d)\n",item);
#endif

    switch (item)
    {
        case itemEditStart:
            Edit_Startup_Items();
            break;
            
        case itemNewItem:
            printf("How did you do that!! -not yet available.\n");
            break;
            
        case itemEditItems:
            Edit_LaunchMenu_Items();
            break;
            
        case itemxinitrc:
            RunX11App(myPath, myPathLen,"xinitrc",7,"",0);
            break;
            
        case itemxterm:
            RunX11App(myPath,myPathLen,"xterm",5, "-ls -name `/bin/hostname`",25);
            break;
        
        case itemxedit:
            RunX11App(myPath,myPathLen,"xedit",5,"",0);
            break;
        
        case itemxbiff:
            RunX11App(myPath,myPathLen,"xbiff",5,"-geometry 50x50-0+20 -file /var/mail/`/usr/bin/whoami`",54);
            break;
        
        case itemxclock:
            RunX11App(myPath,myPathLen,"xclock",6,"-geometry 70x70-0-0",19);
            break;
        
        case itemxeyes:
            RunX11App(myPath,myPathLen,"xeyes",5,"-geometry 100x60-50+20",22);
            break;
        
        case itemxcalc:
            RunX11App(myPath,myPathLen,"xcalc",5,"",0);
            break;
            
        default:
            if(item>itemUserItems)
                RunX11Script(myPath,myPathLen,item);
            else
                printf("What menu item was that? %d\n",item);
            break;
    }        
}

/*
 * Launch an item from the Launch menu given its name
 */
 
int LaunchMenuItem(char *itemname, int hasext, int hasnl)
{
    MenuRef	menu;
    Str255	itemStr;
    int		namelen, item, itemlen;
    pstr_t	*pStr;
    char	*cStr;

    itemlen=strlen(itemname);
    
//#ifdef SHOW_TRACE2
    printf("TRACE2: LaunchMenuItem (itemname,itemlen,hasext: %s , %d,%d)\n",itemname,itemlen,hasext);
//#endif

// Strip off the .x11app extension, if necessary
    if(hasext!=0) {
        if(itemlen<8) return -2;
        itemlen-=7; }
        
// And strip off the newline, if necessary
    if(hasnl!=0) {
        if(itemlen<2) return -2;
        itemlen-=1; }
    
    menu = GetMenuHandle( menuLaunch );

// Loop through the items in the Launch menu, looking for a match
    for(item=itemxinitrc;item<=launchItemsCount;item++) {
        GetMenuItemText(menu,item,itemStr);
    
// use that nasty hack again, to get the string out of the pascal string
// - there's probably a macro to do this 'legally' somewhere!
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
                
// Now compare the strings, and if they match Launch that item
        if(namelen==itemlen)
            if(!strncmp(cStr,itemname,itemlen)) {
                LaunchItem(item);
                return 0; }
    }
    
// failed to find item
    return -1;
}

void DoMenuCommand(long menuResult)
{
    short	menuID;		/* the resource ID of the selected menu */
    short	menuItem;	/* the item number of the selected menu */
	
#ifdef SHOW_TRACE1
    printf("TRACE1: DoMenuCommand (menuResult: %#lx)\n",menuResult);
#endif

    menuID = HiWord(menuResult);    /* use macros to get item & menu number */
    menuItem = LoWord(menuResult);
	
#ifdef SHOW_EVENTS
    printf("Got menuID, menuItem: %d, %d (%#lx)\n",menuID,menuItem,menuResult);
#endif
    switch (menuID) 
    {
        case menuApplication:
            switch (menuItem) 
            {
                case itemAbout:
                    DoAboutBox();
                    break;
                    
/*                case itemQuit:
                    quit();
                    system("/bin/sleep 1");
                    ExitToShell();
                    break;
*/				
                default:
                    break;
            }
            break;
        
        case menuLaunch:
            LaunchItem(menuItem);
            break;
            
        case menuWindow:
//            Restart_XDarwin();
            manipulate_focus_client(menuItem);
//            Stop_XDarwin();
            break;
		
        case menuEdit:
            break;

        case menuOptions:
//            Restart_XDarwin();
            change_options(menuItem);
//            Stop_XDarwin();
            break;

        case menuHelp:
            if(menuItem==itemHelp)
                ShowOroborOSXHelp();
            break;
    }
    HiliteMenu(0);	// unhighlight what MenuSelect (or MenuKey) hilited
//    Restart_XDarwin();
    force_ungrab();	// Try to force ungrabbing of cursor (sometimes gets grabbed on menu select)
//    Stop_XDarwin();
}

void DrawWindow(WindowRef window)
{
    Rect		tempRect;
    GrafPtr		curPort;
	
#ifdef SHOW_TRACE3
    printf("TRACE3: DrawWindow (window: %#lx)\n",window);
#endif

    GetPort(&curPort);
    if(window!=aboutWin)
        MoveWindow(window,-5000,-5000,false);
    SetPort(GetWindowPort(window));
    BeginUpdate(window);
    EraseRect(GetWindowPortBounds(window, &tempRect));
    DrawControls(window);
    DrawGrowIcon(window);
    EndUpdate(window);
    SetPort(curPort);
}

void DoAboutBox(void)
{
#ifdef SHOW_TRACE3
    printf("TRACE3: DoAboutBox\n");
#endif

//   (void) Alert(kAboutBox, nil);  / simple alert dialog box
    aboutWin=(WindowRef) MakeWindow("About OroborOSX",aboutWintype);
    if(aboutWin==NULL) return;
    
    gX11InFront=0;
    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0);
    gone_to_back();

    if(lastFrontWin=noFocusWin) {
        HideWindow(noFocusWin);
        showingNoFocusWin=1;
    }
        
    ShowWindow(aboutWin);
    SelectWindow(aboutWin);
#ifdef DEBUG
    printf("Showing about window: %#lx\n",aboutWin);
#endif
    showingAboutWin=1;
    lastFrontWin=aboutWin;
}

void ShowOroborOSXHelp(void)
{
// Just use the About box for now...
    DoAboutBox();
}
